from distutils.core import setup
setup(
    name = 'lggr',
    packages = ['lggr'],
    version = '0.1.5',
    description = 'Python Logging For Humans',
    author = 'Peter Downs',
    author_email = 'peterldowns@gmail.com',
    url = 'https://github.com/peterldowns/lggr',
    download_url = 'https://github.com/peterldowns/lggr/tarball/v0.1.4',
    keywords = ['logging', 'log', 'coroutine'],
    classifiers = [
        'Programming Language :: Python',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Topic :: System :: Logging',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
)
